#Python device communications protocol framework (pydcpf)
#Copyright (C) 2013  Ondřej Grover
#
#pydcpf is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#pydcpf is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with pydcpf.  If not, see <http://www.gnu.org/licenses/>.
"""This module implements a Device class for communication in Spinel 97 protocol with AD4ETH, AD4RS, AD4USB and Drak 4 devices made by Spinel s.r.o.
"""
import struct

from .. import core

class Device(core.Device):
    """Class representing a AD4ETH, AD4RS, AD4USB and Drak 4
    device made by Spinel s.r.o.

    The Spinel 97 protocol is used for communication
    """

    def __init__(self, address, **kwargs):
        super(Device, self).__init__(address, protocol_module='pydcpf.protocols.spinel97', **kwargs)


    def get_inputs_measured_value_state(self, address=0xfe):
        """Return the values measured and the state of the 4 channels

        Returns
        -------
        channel_measured_values : list of [int, [bool,bool,bool], int]
            list of 4 lists describing each channel:
                channel_number: int
                [underflow, overflow, valid]: list of bool
                value: int
        """
        data = self.query(INST='\x51', DATA='\x00', ADR=address) # DATA for future compatibility according to docs
        channels = []
        for i in xrange(0, 4):          # for all 4 channels
            offset = i*4
            channel_nr, status, value = struct.unpack_from(">cBH", data[offset:offset + 4])
            channels.append([ord(channel_nr),
                             # [ underflow, overflow, valid] ... [3. bit, 4. bit, 8. bit]
                             # in docs bits are indexed form 0, so there it is 2.,3.,7.
                             [bool(status & 8), bool(status & 16), bool(status & 128)],
                             value,
                             ]
                            )
        return channels


    def get_inputs_measured_value(self, address=0xfe, *channels):
        """Return the last measured value by the specified channels

        This method checks the status of the reported values

        Parameters
        ----------
        channels: *args, int
            channels to return

        Returns
        -------
        values : list of int
            last measured values by the given channels

        Raises
        ------
        ValueError: if underflow, overflow or invalid value detected or wrong channel reported
        """
        channels_state = self.get_inputs_measured_value_state(address)
        values = []
        for channel in channels:
            channel_nr, (underflow, overflow, valid), value = channels_state[channel - 1]
            if channel_nr == channel and valid and not underflow and not overflow:
                values.append(value)
            elif channel_nr != channel:
                raise ValueError("channel %i reported channel number" % (channel, channel_nr))
            elif underflow:
                raise ValueError("channel %i reported underflow" % channel)
            elif overflow:
                raise ValueError("channel %i reported overflow" % channel)
            elif not valid:
                raise ValueError("channel %i reported invalid value" % channel)
        return values
                             

    def get_real_measured_values_state(self, address=0xfe, *channels):
        """Return a data structure for each given channel describing the real measured value

        Returns
        -------
        output : list of [int, [bool,bool,bool], float, int, str]
        """
        if channels == []:
            DATA = '\x00'
        else:
            DATA = [ chr(channel) for channel in channels ]
            DATA = "".join(DATA)
        data = self.query(INST='\x58', DATA=DATA, ADR=address)
        output = []
        for (i, channel) in enumerate(channels):
            offset = i*18
            channel_nr, status, int_value, float_value, str_value = struct.unpack_from(">cBHf10s", data[offset:offset + 18])
            output.append([ord(channel_nr),
                             # [ underflow, overflow, valid] ... [3. bit, 4. bit, 8. bit]
                             # in docs bits are indexed form 0, so there it is 2.,3.,7.
                             [bool(status & 8), bool(status & 16), bool(status & 128)],
                             float_value, str_value, int_value
                             ]
                            )
        return output
    



#Python device communications protocol framework (pydcpf)
#Copyright (C) 2013  Ondřej Grover
#
#pydcpf is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#pydcpf is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with pydcpf.  If not, see <http://www.gnu.org/licenses/>.
from .. import core


class Device(core.Device):


    def __init__(self, address, **kwargs):
        """Initialize a device communicating with the EVR116 valve

        address is expected to be a socket address of the RS232ETH converter or the
        or path of the serial port character file node

        Serial line settings: 300 Baud, 7 bits, 2 stop bits, log. 0 >7V, log. 1 <3V 
        """
        super(Device, self).__init__(address=address, protocol_module='pydcpf.protocols.evr116', **kwargs)

    def set_position(self, position):
        if position < 512 or position > 6760:
            raise ValueError("position not in range [512, 6760]")
        hex_position = "%x" % (position / 2)
        return self.query(IDENTIFIER='g', DATA=hex_position)


    def get_position(self):
        hex_position = self.query(IDENTIFIER='p', DATA='?')
        return int(str(hex_position), 16)


    def close_valve(self):
        return self.query(IDENTIFIER='x')


    def open_valve(self):
        return self.query(IDENTIFIER='y')

    



#Python device communications protocol framework (pydcpf)
#Copyright (C) 2013  Ondřej Grover
#
#pydcpf is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#pydcpf is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with pydcpf.  If not, see <http://www.gnu.org/licenses/>.
from .spinelbase import SpinelBasePacket, ACKError


__all__ = ["RequestPacket", "ResponsePacket"]




class CheckSumError(Exception):

    
    def __init__(self, packet):
        self.packet = packet


    def __str__(self):
        return "Packet checksum is 0x%02x, but the SUMA checksum byte is 0x%02x" % (self.packet.calculate_checksum(), self.packet.SUMA)




class Spinel97BasePacket(SpinelBasePacket):

    
    def calculate_checksum(self):
        end = self.start + self.length - 2
        SUM = 255
        packet = self.raw_packet
        for i in xrange(self.start, end):
            SUM -= packet[i]
        return abs(SUM % 256)

    
    def check(self):
        if self.SUMA != self.calculate_checksum():
            raise CheckSumError(self)
        try:
            if self.ACK != '\x00':
                raise ACKError(self)
        except AttributeError:
            pass # must be a RequestPacket which has INST instead of ACK


    def find(self):
        sup = super(Spinel97BasePacket, self)
        full_buffer_length = len(self.raw_packet)
        if sup.find(): #means we have a candidate at all
            while full_buffer_length - self.start >= 9:
                # first check whether it could be a valid packet
                # based on the count of necessary bytes in packet
                packet_len = self.NUM + 4
                CR_position = self.start + packet_len - 1 #CR_position must be an index
                if CR_position < full_buffer_length and self.raw_packet[CR_position] == 13:
                    # if the last byte is CR as reported by NUM, ord('\r') == 13
                    self.length = packet_len
                    return True
                if not sup.find(self.start + 1): #look for another candidate
                    return False
        else:
            return False
            


Spinel97BasePacket.register_element('NUM', 'Number of bytes in packet after NUM', start_position=2, code='>H')
Spinel97BasePacket.register_element('ADR', 'Module address number', start_position=4, code='>B')
Spinel97BasePacket.register_element('SIG', 'Signature number', start_position=5, code='>B')
Spinel97BasePacket.register_element('DATA', 'Data contained in packet', start_position=7, end_position=-2)
Spinel97BasePacket.register_element('SUMA', 'Checksum number', start_position=-2, code='>B')




class RequestPacket(Spinel97BasePacket):
    """RequestPacket class for the Spinel 97 protocol format"""

    
    def __init__(self, INST=None, ADR=0xfe, DATA='', NUM=None, SIG=None, SUMA=None, raw_packet=None ):
        if raw_packet is not None:
            super(RequestPacket, self).__init__(raw_packet=raw_packet)
        elif INST is not None:
            super(RequestPacket, self).__init__(PRE='*', FRM=97, ADR=ADR, INST=INST, DATA=DATA, CR='\r')
            if NUM is None:
                NUM = self.length - 4
            self.NUM = NUM
            if SIG is None:
                SIG = 2
            self.SIG = SIG
            if SUMA is None:
                SUMA = self.calculate_checksum()
            self.SUMA = SUMA



RequestPacket.register_element('INST', "Device instruction code character", start_position=6, length=1)



class ResponsePacket(Spinel97BasePacket):
    """ResponsePacket class for the Spinel 97 protocol format"""

    
    def __init__(self, ACK=None, ADR=0xfe, DATA='', NUM=None, SIG=None, SUMA=None, raw_packet=None ):
        if ACK is not None or raw_packet is not None:
            super(ResponsePacket, self).__init__(raw_packet=raw_packet, PRE='*', FRM=97, ADR=ADR, ACK=ACK, DATA=DATA, CR='\r')
            if NUM is None:
                NUM = self.length - 4
            self.NUM = NUM
            if SIG is None:
                SIG = 2
            self.SIG = SIG
            if SUMA is None:
                SUMA = self.calculate_checksum()
            self.SUMA = SUMA



ResponsePacket.register_element('ACK', 'Acknowledgment code character', start_position=6, length=1)




#Python device communications protocol framework (pydcpf)
#Copyright (C) 2013  Ondřej Grover
#
#pydcpf is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#pydcpf is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with pydcpf.  If not, see <http://www.gnu.org/licenses/>.
from .. import core


class Device(core.Device):
    """Device representing a AC250Kxxx series power supply by Diametral

    Communications goes over a serial line with a baudrate of 9600, bytesize of 8, no bit parity, 1 stopbit and no flow control
    """


    def __init__(self, serial_port, address=0xff, **kwargs):
        """Initialize a new Device object with the specified address communicating through the specified serial port.

        Parameters
        ----------
        serial_port : int or str
            number of the serial port to be used for communication
            or the explicit name of the device to be passed to :meth:`serial.Serial.__init__`
            on Linux it's the number X in /dev/ttySX and defaults to 0 (/dev/ttyS0)
        address : int, optional
            internal device address
            ranges from 0 to 31 inclusive, defaluts to 255 (broadcast address)
            address 255 is the broadcast address, any device accepts it, but won't send a response packet back (not recommended, makes debugging very hard)
            the device address can be displayed by holding the red 'Clear' button for several seconds
        kwargs : keyword arguments dict
            passed on to :meth:`core.Device.__init__`
        """
        self.internal_address = address
        super(Device, self).__init__(address=serial_port, interface_module="pydcpf.interfaces.serial_interface",
                                     protocol_interface="pydcpf.protocols.AC250Kxxx",
                                     baudrate=9600, bytesize=8, parity='N', stopbits=1, xonxoff=False)


    def query(self, send_byte_count=None, receive_byte_count=None, check_parameters=dict(), **packet_parameters):
        """Wrapper around :meth:`core.Device.query` taht inserts the internals address if none is specified"""
        if not "ADR" in packet_parameters:
            packet_parameters["ADR"] = self.internal_address
        super(Device, self).query(send_byte_count, receive_byte_count, check_parameters, **packet_parameters)
        
            
    def command(self, send_byte_count=None, receive_byte_count=None, check_parameters=dict(), **packet_parameters):
        """Device.command(instruction) -> ack

        Send a command to the device and wait for acknowledgment (ACK)
        Essentially just a wrapper around :meth:`Device.query`

        Parameters
        ----------
        same as for :meth:`Device.query`

        Returns
        -------
        ack : bool
            True if Device replied 'OK'
            False if Device replied 'Err'
            
        Raises
        ------
        :class:`RuntimeError`
            Raises if the reply was something else than 'OK' or 'Err'
        """
        ack = self.query(send_byte_count, receive_byte_count, check_parameters, **packet_parameters)
        if ack == 'OK':
            return True
        elif ack == 'Err':
            return False
        else:
            raise RuntimeError("Device reported error: " + ack)

        
    def get_voltage(self):
        """Device.get_voltage() -> voltage

        Return the current set voltage in Volts

        Returns
        -------
        voltage : int
            current voltage in Volts
        """
        return int(self.query(DATA='NAP???')[3:]) #reply is 'NAPXXX'

        
    def set_voltage(self, voltage):
        """Device.set_voltage(voltage) -> success

        Set the voltage in Volts

        Parameters
        ----------
        voltage : int
            voltage to set in Volts

        Returns
        -------
        success : bool
            True if the command did succeed, False otherwise
        """
        return self.command(DATA=('NAP%03d' % voltage)) #it takes some time for the voltage to change

        
    voltage = property(fget=get_voltage, fset=set_voltage, doc="""Output voltage as an integer in Volts""")

    
    def get_output(self):
        """Device.get_output() -> status

        Return the current status of the output

        Returns
        -------
        status : bool
            True if output is activated, False otherwise
        """
        if self.query(DATA='OUT?')[-1] == '1': #should be 'OUT1'
            return True
        else: #should be 'OUT0'
            return False

        
    def set_output(self, status):
        """Device.set_output(status) -> success

        Set the status of the output

        Parameters
        ----------
        status : bool
            True if output should be activated, False otherwise
        
        Returns
        -------
        success : bool
            True if the command did succeed, False otherwise
        """
        if status: #if True
            return self.command(DATA='OUT1')
        else:
            return self.command(DATA='OUT0')

        
    output = property(fget=get_output, fset=set_output, doc="""Output status as a Boolean, True if activated, False otherwise""")
    
        
    def get_identification(self):
        """Device.get_identification() -> identification

        Return the identification of the device

        Returns
        -------
        identification : str
            name of the device, model and revision
        """
        return self.query(DATA='ID?')

    identification = property(fget=get_identification, doc="""Device identifiaction as a string""")
        



#Python device communications protocol framework (pydcpf)
#Copyright (C) 2013  Ondřej Grover
#
#pydcpf is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#pydcpf is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with pydcpf.  If not, see <http://www.gnu.org/licenses/>.
"""This module provides the base :class:`Interface` class for documentation purposes."""

__all__ = ['Interface']


class Interface(object):
    """Base class for device data transmission"""

    
    def __init__(self, timeout, **kwargs):
        """Initialize the interface using the keyword arguments (implementation dependent).

        The timout argument represents a timout in seconds for connecting, reading writing. If 0, no timeout should be set. The timeout error raised on timout is implementation dependent.

        The interface must not connect immediately after initialization, :meth:`Interface.connect` must be used for that.
        """
        pass


    def connect(self, address, serve):
        """Connect the interface to the address specified, possibly in a special way if serving.
        
        Parameters
        ----------
        address
                The type and content depends on the implementation.
        server : bool
            May modify the behavior of the method depending on the implementation, e.g. if True, :meth:`Interface.connect` may wait for a connection on *address*, if False, it may request a connection from *address*.
            For simpler interfaces, the server and client behavior may be the same.      
        """
        pass


    def disconnect(self, address, serve):
        """Disconnect the interface, possibly in a special way if serving.
        
         Parameters
        -----------
        server : bool
            May modify the behavior of the method depending on the implementation, e.g. if True, :meth:`Interface.disconnect` may take a few more action to clean up the interface properly.
            For simpler interfaces, the server and client behavior may be the same.      """
        pass
    

    def send_data(self, data):
        """Sends the data through the interface.

        Parameters
        ----------
        data : str or bytearray or bytes
            data to be sent

        Note
        ----
        This method is expected to send *ALL* data, so it may have to check whether all data was sent
        """
        pass
    

    def receive_data(self, byte_count):
        """Receive up to byte_count bytes from the interface and return them

        Parameters
        ----------
        byte_count : int
            number of bytes to attempt to read, may return equal or less

        Return
        ------
        data : str or bytearray, or bytes
            received data
        """
        pass
    
        



#Python device communications protocol framework (pydcpf)
#Copyright (C) 2013  Ondřej Grover
#
#pydcpf is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#pydcpf is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with pydcpf.  If not, see <http://www.gnu.org/licenses/>.
"""API for the Quido IO module made by Papouch s.r.o."""

from .. import core
import struct 


_outputs_inputs_count_fmts = {
    1 : 'B',
    2 : 'H',
    4 : 'I',
    }


class Device(core.Device):
    """Class representing a connected Quido module

    Communication is done through the Spinel 97 protocol,
    becuse it's more reliable than the Spinel 66 protocol
    """


    def __init__(self, address, **kwargs):
        super(Device, self).__init__(address, protocol_module='pydcpf.protocols.spinel97', **kwargs)


    def _decode_inputs_outputs_state(self, data):
        """Common code for decoding responses
        describing inputs or outputs state

        Note
        ----
            only for max 32 inputs/outputs
        """
        data_len = len(data)
        fmt = _outputs_inputs_count_fmts[data_len]
        state = [ bool(int(i)) for i in
              ("%0" + "%ii" % 8**data_len) % int(bin(struct.unpack_from(fmt, data)[0])[2:]) # crunch down to padded binary string representation
              ]
        state.reverse()        # reverse in place as specified in docs
        return state


    def get_outputs_state(self, address):
        """Return the state of all outputs as a list
        on the device with the specified address

        Returns
        -------
        outputs_state : list of bool
            length depends of number of outputs on device
            True if active (high voltage)
            False if inactive (low voltage)
        """
        return self._decode_inputs_outputs_state(
            self.query(INST='\x30', ADR=address)
            )

        
    def get_output_state(self, output_number, address):
        """Return the state of output given by *output_number*
        on the device with the specified address

        Returns
        -------
        output_state : bool
            

        Note
        ----
        This method has to call :meth:`Device.get_outputs_state`
        to get the status of all outputs first anyways,
        so it's more resourceful to batch output querying
        """
        return self.get_outputs_state(address)[output_number - 1] # list indexed from 0
    

    def set_outputs_state(self, address, *outputs_state):
        """
        Set each specified output to the given state

        Parameters
        ----------
        outputs_state : list of ints
            output numbers to operate on,
            set active (high voltage) if positive,
            inactive (low voltage) if negative
            no order is required, arbitrary outputs can be given
        """
        return self.query(
            INST='\x20',
            DATA=struct.pack("%ib" % len(outputs_state),
                             # 1 bit H/L (-/+ bit) + 7 bits for output number from 1 to 127
                             *[ abs(i) if i < 0 else i - 128 for i in outputs_state ]),
            )


    def get_inputs_state(self, address):
        """Return the state of all inputs as a list
        on the device with the specified address

        Returns
        -------
        inputs_state : list of bool
            length depends of number of inputs on device
            True if active (high voltage)
            False if inactive (low voltage)
        """
        return self._decode_inputs_outputs_state(
            self.query(INST='\x31', ADR=address)
            )



#Python device communications protocol framework (pydcpf)
#Copyright (C) 2013  Ondřej Grover
#
#pydcpf is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#pydcpf is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with pydcpf.  If not, see <http://www.gnu.org/licenses/>.
from . import base
import socket


class Interface(base.Interface):
    """Wrapper class around :class:`socket.socket`.

    Wrapping it is necessary because after disconnecting a new socket must be created when connecting again a nad also when serving

    Raises :class:`socket.timeout` error on timeout"""


    def _create_socket(self):
        self.socket = socket.socket(*self._socket_parameters)
        self.socket.settimeout(self._timeout)
    
    def __init__(self, timeout, family=socket.AF_INET, type=socket.SOCK_STREAM, protocol=0, _sock=None):
        self._socket_parameters = (family, type, protocol, _sock) # needed in disconnecting
        self._timeout = timeout
        self._create_socket()
        
    def connect(self, address, serve):
        if serve:
            self.socket.bind(address)
            self.socket.listen()
            self.socket = self.socket.accept()[0]
        else:
            self.socket.connect(address)
            

    def disconnect(self, address, serve):
        self.socket.close()
        self._create_socket()

    def send_data(self, data, flags=0):
        self.socket.sendall(data, flags)
        

    def receive_data(self, byte_count, flags=0):
        return self.socket.recv(byte_count, flags)



#Python device communications protocol framework (pydcpf)
#Copyright (C) 2013  Ondřej Grover
#
#pydcpf is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#pydcpf is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with pydcpf.  If not, see <http://www.gnu.org/licenses/>.
"""This package provides the base :class:`base.Interface` class and several other Interface classes in submodules
"""
from base import Interface
import base
__all__ = base.__all__



#Python device communications protocol framework (pydcpf)
#Copyright (C) 2013  Ondřej Grover
#
#pydcpf is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#pydcpf is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with pydcpf.  If not, see <http://www.gnu.org/licenses/>.
from . import base


def _hexify(value):
    """trasform a number into a 1 Byte hex representation suitable for communication"""
    return "%02X" % value


def _dehexify(hexbytestring):
    """Transform a 2-char bytestring representing a number"""
    return int(str(hexbytestring), 16)


class _basePacket(base.RequestPacket):


    def _set_address(self, address):
        self.raw_packet[1:3] = _hexify(address)

    def _get_address(self):
        return _dehexify(self.raw_packet[1:3])


_basePacket.register_element("INIT", "initializing character", start_position=0, length=1)
_basePacket.register_element("ADR", "device address in decimal", set_function=_basePacket._set_address, get_function=_basePacket._get_address, length=2)
_basePacket.register_element("CR", "terminating character", start_position=-1, length=1)

    

class RequestPacket(_basePacket):


    def __init__(self, ADR=10, DATA=''):
        super(RequestPacket, self).__init__(INIT='@', ADR=ADR, DATA=DATA, CR='\r')
        _sum = 0
        for i in self.raw_packet[1:-3]:
            _sum += i
        while _sum > 256: #the sum must be lesser or equal to 256
            _sum -= 256
        self.CTRLSUM = _sum


    def _set_ctrlsum(self, value):
        self.raw_packet[-3:-1] = _hexify(value)


    def _get_ctrlsum(self):
        return _dehexify(self.raw_packet[-3:-1])


    def find(self):
        pass


RequestPacket.register_element("DATA", "data contained in packet", start_position=3, end_position=-3)
RequestPacket.register_element("CTRLSUM", "control sum in decimal", set_function=RequestPacket._set_ctrlsum, get_function=RequestPacket._get_ctrlsum, length=2)



class ResponsePacket(_basePacket):


    def find(self):
        raw_packet = self.raw_packet #minimize attr lookups
        try:
            start = raw_packet.index('#')
            end = raw_packet.index('\r', start)
        except ValueError:
            return False
        self.start, self.length = start, end - start + 1
        return True


import copy
#to prevent conflicts with previous DATA definition
ResponsePacket.elements_definitions_dict = copy.copy(ResponsePacket.elements_definitions_dict)
ResponsePacket.register_element("DATA", "data contained in packet", start_position=3, end_position=-1)



#Python device communications protocol framework (pydcpf)
#Copyright (C) 2013  Ondřej Grover
#
#pydcpf is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#pydcpf is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with pydcpf.  If not, see <http://www.gnu.org/licenses/>.
from .base import ResponsePacket



class ACKError(Exception):
    """Acknowledgment code error

    Attributes
    ----------
    ACK : str
        Acknowledgment code in the packet
    """

    descriptions = {
            0x00 : "OK",
            0x01 : "Unknown error",
            0x02 : "Invalid instruction",
            0x03 : "Invalid instruction parameters",
            0x04 : "Permission denied", #write error, too low data, channel disabled, other requirements not met
            0x05 : "Device malfunction",
            0x06 : "Data not available",
            0x0d : "Digital input state change", #automatically sent message
            0x0e : "Continuous measurement", #automatically sent, repeatedly sending measured data
            0x0f : "Range overrun" #automatically sent
        }

    
    def __init__(self, packet):
        self.packet = packet

        
    def __str__(self):
        try:
            int_ACK = int(self.packet.ACK, 16)
        except ValueError: #must be something like '\x0d', so use ord
            int_ACK = ord(self.packet.ACK)
        try:
            return "Non-zero acknowledgment code 0x%02x: " %  int_ACK + self.__class__.descriptions[int_ACK]
        except KeyError: #unknow error code
            return "Unknown acknowledgment code 0x%02x" % int_ACK


    
class SpinelBasePacket(ResponsePacket):


    def find(self, buffer_start=0):
        raw_packet = self.raw_packet #minimize attr lookups
        try:
            start = raw_packet.index('*', buffer_start)
            end = raw_packet.index('\r', start)
        except ValueError:
            return False
        self.start, self.length = start, end - start + 1
        return True



SpinelBasePacket.register_element('PRE', "Packet prefix character", start_position=0, length=1)
SpinelBasePacket.register_element('FRM', "Packet format number", start_position=1, code='>B')
SpinelBasePacket.register_element('CR', "Ending mark character", start_position=-1, length=1)



#Python device communications protocol framework (pydcpf)
#Copyright (C) 2013  Ondřej Grover
#
#pydcpf is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#pydcpf is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with pydcpf.  If not, see <http://www.gnu.org/licenses/>.

"""API for the DAS1210 device made by Paouch s.r.o. used at the GOLEM reactor, FJFI, CVUT"""

from . import spinel_core
import struct


ranges = [0.25, 0.5, 1, 2.5, 5, 10]


class Device(spinel_core.Device):
    """Class representing the DAS1210 device

    Methods
    -------
    Almost all methods have a optional channel parameter
        this parameter determines the channel number ot operate on
        or (default value) operates on all channels otherwise (using the universal address).
    Where possible, the method parameters default to the values that are set on device reset.
    Methods beginning with 'set_'  return the DATA portion of the packet, usually empty
    Methods beginning with 'get_' return some meaningful value, see their docstring for more.
    """


    def __init__(self, ip_address, port=10001, **kwargs):
        """Initalize the device

        Parameters
        ----------
        ip_address : str
            IPv4 address of the device
        port : int, optional
            TCP port on which to initiate communication
            default is 10001
        **kwargs : keyword arguments
            these are passed on to :meth:`core.Device.__init__`
        """
        super(Device, self).__init__(address=(ip_address, port), protocol_module='pydcpf.protocols.spinel97', interface_module='pydcpf.interfaces.socket_interface', **kwargs)
    
    
    def set_range(self, value=10, channel=0xff):
        """Set the channel range.

        Parameters
        ----------
        value : int
            digitization range amplitude
            possible values [V]: 0.25, 0.5, 1, 2.5, 5, 10
            resulting range is from -value to +value
            
        Raises
        ------
        ValueError
            if specified range cannot be set (invalid value)
        """
        
        try:
            return self.query(INST='\x70', DATA=chr(ranges.index(value)), ADR=channel)
        except ValueError:
            raise ValueError("Invalid range specified, possible ranges are: " + repr(ranges))

        
    def get_range(self, channel):
        """Return the range of the channel in Volts.
        
        Inverse function to :func:`set_range`
        """
        try:
            return ranges[ord(str(self.query(INST='\x71', ADR=channel)))]
        except IndexError:
            raise ValueError("Device reports unexpected range with identifier " + value)

 
    def set_trigger(self, trigger=True, channel=0xff):
        """Set the trigger mode of the specified channel.

        Parameters
        ----------
        trigger : bool
            if True, use a rising edge trigger mode
            if False, use falling edge trigger mode
        """
        return self.query(INST='\x72', DATA=chr(int(trigger)), ADR=channel)


    def get_trigger(self, channel):
        """Return the trigger mode as bool as described in the inverse function :func:`set_trigger`
        """
        return bool(ord(str(self.query(INST='\x73', ADR=channel))))

    
    def set_sampling_frequency(self, freq=1e6, channel=0xff):
        """Set the sampling frequency of the specified channel.

        Parameters
        ----------
        freq : float
            freq for the specified channel in Hz
            ranges from 39.0625 kHz ('\xff') to 1.25 MHz ('\x07')
            default is 1 MHz
            the possible frequencies are given by the formula freq = 1e7 / (byte + 1)
            where byte ranges from '\x07' to '\xff'
            Therefore, the specified freq is rounded down to the nearest possible value.
        """
        return self.query(INST='\x74', DATA=chr(int(1e7 / freq) - 1), ADR=channel)


    def get_sampling_frequency(self, channel):
        """Return the sampling frequency for the specified channel in Hz as a float
        """
        return 1e7 / (ord(str(self.query(INST='\x75', ADR=channel)) + 1))


    def set_samples_count(self, count=524287, channel=0xff):
        """Set the number of samples to record by the specified channel.
        The actual number is rounded up to the closest multiple of 8.

        Parameters
        ----------
        count : int
            number of samples to set
            ranges from 0 to 524287 (default)
        """
        return self.query(INST='\x76', DATA=struct.pack('>i', count), ADR=channel)


    def get_samples_count(self, channel):
        """Return the number of samples that are recorded by the specified channel.
        Inverse function to :func:`set_samples_count`
        """
        return struct.unpack_from('>i', self.query(INST='\x77', ADR=channel))[0]


    def get_data(self, length, channel, packet_size=4096):
        """Retreive a data sample of the specified length from the specified channel.

        Parameters
        ----------
        length : int
            number of data points in sample
            should be a multiple of *packet_size*
        channel : int
            channel number
        packet_size : int
            number of data points to retreive in one packet
            defaults to 4096 which is the optimum value with regards to the device IO capabilities
            must not exceed 8192
            
        Returns
        -------
        data_array : list of buffers
            
        """
        target = [] #will append to the list
        packet_size_raw = struct.pack('>i', packet_size) #this needs to be done only once
        for packet_i in xrange(length / packet_size): #number of packets needed, rounded up to inlude the requested length for sure
            target.append(self.query(INST='\x51', DATA=struct.pack('>i', packet_i * packet_size) + packet_size_raw, ADR=channel))
        return target

    
    def get_data_ready(self, channel):
        """Return True if data are ready, False otherwise
        """
        return bool(ord(str(self.query(INST='\xf5', ADR=channel))))


    def set_ready(self, channel=0xff):
        """Set the specified channel operational.
        This must be done after the channel has recorded some data after trigger to empty the data memory buffer for the next measurement.
        """
        return self.query(INST='\x78', ADR=channel)


    def get_version(self):
        """Return a string describing the device and its version
        """
        return str(self.query(INST='\xf3', ADR=1))



#Python device communications protocol framework (pydcpf)
#Copyright (C) 2013  Ondřej Grover
#
#pydcpf is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#pydcpf is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with pydcpf.  If not, see <http://www.gnu.org/licenses/>.
from . import base
import serial



class Interface(base.Interface, serial.Serial):
    """Extended class of :class:`serial.Serial`.

    It only translates some of the method names and takes care of connecting on demand
    """


    def __init__(self, timeout, **kwargs):
        kwargs["timeout"] = timeout
        try:
            del kwargs["port"] #make sure the interface does not connect immediately
        except KeyError:
            pass                          #if port was not set, no problem
        super(Interface, self).__init__(**kwargs)
    
        
    def connect(self, address, serve):
        self.setPort(address)
        self.open()

        
    def disconnect(self, address, serve):
        self.close()

        
    def send_data(self, data):
        self.write(data)
        

    def receive_data(self, byte_count):
        return self.read(byte_count)



#Python device communications protocol framework (pydcpf)
#Copyright (C) 2013  Ondřej Grover
#
#pydcpf is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#pydcpf is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with pydcpf.  If not, see <http://www.gnu.org/licenses/>.
"""This module provides the base :class:`RequestPacket` and :class:`ResponsePacket` classes, mainly for documentation purposes (some methods shouldn't to be overridden).
"""

__all__ = ["RequestPacket", "ResponsePacket"]

import struct



class RequestPacket(object):
    """Request packet class

    This class should be used when sending requests to the device

    For most protocols this class may be the same as :class:`ResponsePacket`
    """
    
    
    def __init__(self, raw_packet=None, **packet_parameters):
        """Initialize the packet either from the *raw_packet* or from keyword arguments or just simply create it if nothing is specified

        Parameters
        ----------
        raw_packet : int or bytearray, optional
            If specified, it will be stored in :attr:`RequestPacket.raw_packet`
            If int, an empty bytearray of that size will be made 
        **packet_parameters, optional
            keyword parameters used for constructing the packet
            each keyword argument should be the name of a property defined with :meth:`RequestPacket.register_element`

        Note
        ----
        This method should NOT be completely overridden, define your own packet structure using :meth:`RequestPacket.register_element` instead or remember to call it at the end of your own __init__!
        """
        if raw_packet is not None:
            if isinstance(raw_packet, int):
                raw_packet = bytearray(raw_packet)
            self.raw_packet = raw_packet
            self.start, self.length = 0, len(raw_packet)
        elif packet_parameters != {}:
            minimum_packet_length = self.__class__.minimum_packet_length
            for name in packet_parameters.iterkeys():
                if self.__class__.elements_definitions_dict[name][1] is None:
                    minimum_packet_length += len(packet_parameters[name])
            self.raw_packet = bytearray(minimum_packet_length)
            self.start, self.length = 0, len(self.raw_packet)
            for name, value in packet_parameters.iteritems():
                setattr(self, name, value)


    def _get_element_substring(self, name):
        """Return a character or substring representing the named packet element"""
        start_position, length_or_code, end_position = self.__class__.elements_definitions_dict[name]
        start_position += self.start
        if start_position < 0:
            start_position += self.length
        if isinstance(length_or_code, str): #is a code
            return struct.unpack_from(length_or_code, buffer(self.raw_packet), start_position)[0]
        if length_or_code > 1: #must be int then
            return buffer(self.raw_packet, start_position, length_or_code)
        elif length_or_code is not None: #must be 1 then
            return chr(self.raw_packet[start_position])
        if end_position > 0:
            return buffer(self.raw_packet, start_position, end_position - start_position)
        elif end_position is not None: #must be negative then
            return buffer(self.raw_packet, start_position, self.length + end_position - start_position)
        

    def _set_element_substring(self, name, value):
        """Set the named packet element to a character or substring value"""
        start_position, length_or_code, end_position = self.__class__.elements_definitions_dict[name]
        start_position += self.start
        if start_position < 0:
            start_position += self.length
        if isinstance(length_or_code, str): #is a code
            struct.pack_into(length_or_code, self.raw_packet, start_position, value)
        elif length_or_code > 1: #must be int then
            self.raw_packet[start_position:start_position + length_or_code] = value
        elif length_or_code is not None: #must be 1 then
            self.raw_packet[start_position] = value
        elif end_position is not None: 
            self.raw_packet[start_position:end_position] = value
            

    def _del_element_substring(self, name):
        """Delete a character or substring in the raw_packet representing the named packet element"""
        length_or_code = self.__class__.elements_definitions_dict.pop(name)[1]
        if isinstance(length_or_code, str):
            self.__class__.minimum_packet_length -= struct.calcsize(length_or_code)
        elif length_or_code > 0: #so cannot be None
            self.__class__.minimum_packet_length -= length_or_code
            
        
    @classmethod
    def register_element(cls, name, docstring, start_position=None, code=None, end_position=None, set_function=None, length=None, get_function=None, del_function=None):
        """Register an element of the packet structure and make a :class:`property` for accessing a part of the packet with Packet.name  with a *docstring* which will either unpack or pack a value according to the *code* beginning at *start_position* or set or get a buffer starting at *start_position* and ending at *end_position* or provide your own *set_function*, *get_function* and *del_function*

        ::
        class RequestPacket(pydcpf.protocols.base.RequestPacket):
        ...
        RequestPacket.register_element('NUM', 'Number of bytes in the packet after NUM', 2, code='>H')
        RequestPacket.register_element('DATA','Data contained in the packet', 6, end_position=-2)

        Parameters
        ----------
        name : str
            name of the property, will be accessible as Packet.name
        docstring : str
            documentation for the property
        start_position : int, optional
            index in the raw_packet buffer at which to start to pack or unpack the value(s)
        code : str, optional
            C type format string, see :module:`struct` for details
            mutually exclusive with *end_position*
        end_position : int, optional
            index in the raw_packet buffer at which the value ends, in pythonic style
            mutually exclusive with *code*
            may be negative as an index relative to the end
        length : int, optional
            if the byte length cannot be deduced from *code* or *start_position* and *end_position* (e.g. if you specify your own functions or *end_position* is negative) it should be provided so :meth:`RequestPacket.__init__` allocates enough memory, otherwise it is calculated from the length of the provided arguments
        get_function : function, optional
            if not None, it will be used instead of creating one based on *start_position*, *code* or *end_position*
        set_function : function, optional
            if not None, it will be used instead of creating one based on *start_position*, *code* or *end_position*
        del_function : function, optional
            a function that safely deletes and unsets all instance attributes or dictionaries created by this property
            if *code* or *length* was specified, it should not be needed

        Note
        ----
        If you plan on using these properties to set the respective portions of the :attr:`ResponsePacket.raw_buffer`, it is your duty to make sure that the raw_buffer has enough space for the data
        """
        # first make sure the class variables are present
        try:
            cls.minimum_packet_length
        except AttributeError:
            cls.minimum_packet_length = 0
        try:
            cls.elements_definitions_dict
        except AttributeError:
            cls.elements_definitions_dict = {}

        length_or_code = None
        if code is not None:
            cls.minimum_packet_length += struct.calcsize(code)
            length_or_code = code
        elif length is not None:
            length_or_code = length
            cls.minimum_packet_length += length
        elif end_position is not None and end_position > 0:
            length_or_code = end_position - start_position
            cls.minimum_packet_length += length_or_code
        cls.elements_definitions_dict[name] = (start_position, length_or_code, end_position)
        if get_function is None:
            def get_function(self):
                return cls._get_element_substring(self, name)
        if set_function is None:
            def set_function(self, value):
                cls._set_element_substring(self, name, value)
        if del_function is None:
            def del_function(self):
                cls._del_element_substring(self, name)
        setattr(cls, name, property(fget=get_function, fset=set_function, fdel=del_function, doc=docstring))


    def __iter__(self):
        """Return and iterator over (element_name, value) pairs"""
        for name in self.__class__.elements_definitions_dict.iterkeys():
            yield (name, getattr(self, name))
            


class ResponsePacket(RequestPacket):
    """Response packet class

    This class should be used when receiving responses from the device

    For most protocols this class may be the same as :class:`RequestPacket`
    """


    def find(self):
        """Try to find a WHOLE packet in :attr:`ResponsePacket.raw_packet`

        This method must set :attr:`ResponsePacket.start` and :attr:`ResponsePacket.length`
        
        Returns
        -------
        found : bool
            True if a whole packet was found, False otherwise
        """
        pass


    def check(self, **parameters):
        """Check and verify the packet, optionally modify the verification method based on keyword *parameters*

        Parameters
        ----------
        **parameters : implementation dependant, optional
            keyword arguments possibly modifying the verification procedure

        Raises
        ------
        PacketException : implementation dependant
            Each protocol module defines its own exceptions
        """
        pass



#Python device communications protocol framework (pydcpf)
#Copyright (C) 2013  Ondřej Grover
#
#pydcpf is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#pydcpf is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with pydcpf.  If not, see <http://www.gnu.org/licenses/>.
"""This module defines a customized Device class specific to the Spinel protocol

the main difference is that the device will not expect a response when querying a broadcast or universal address
"""

from .. import core


class Device(core.Device):


    def query(self, send_byte_count=None, receive_byte_count=None, check_parameters=dict(), **packet_parameters):
        """Query the device and expect a response based on the address

        According to the Spinel protocol specification,
        if address is the special broadcast or universal address,
        the device will not send a response
        """
        if packet_parameters['ADR'] in [0xff, 0xfe, '%', '$']:
            self.send_request(send_byte_count, **packet_parameters)
        else:
            return super(Device, self).query(send_byte_count, receive_byte_count, check_parameters, **packet_parameters)




#Python device communications protocol framework (pydcpf)
#Copyright (C) 2013  Ondřej Grover
#
#pydcpf is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#pydcpf is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with pydcpf.  If not, see <http://www.gnu.org/licenses/>.

from . import base

valid_command_characters = "hxyzijgstv" # request chars
valid_command_characters += "p" + valid_command_characters.upper() # response chars


class RequestPacket(base.RequestPacket):


    def __init__(self, IDENTIFIER='x', DATA=''):
        super(RequestPacket, self).__init__(IDENTIFIER=IDENTIFIER, DATA=DATA, TERMINATOR='\r\n')

    def check(self):
        pass

    def find(self, buffer_start=0):
        raw_packet = self.raw_packet
        try:
            end = raw_packet.index('\r\n', buffer_start) + 1
        except ValueError:
            return False
        possible_start = raw_packet.rfind('\r\n', buffer_start, end - 1)
        if possible_start == -1:
            possible_start = buffer_start
        else:
            possible_start += 2
        possible_packet_buff = buffer(raw_packet, possible_start, end - possible_start - 1)
        for char in reversed(possible_packet_buff):
            if char in valid_command_characters:
                self.start = raw_packet.rfind(char, possible_start, end - 2)
                self.length = end - self.start + 1
                return True
        # might be a reposnse to 'p?' query - only 4 bytes with no identifier
        if len(possible_packet_buff) == 4:
            self.start = possible_start
            self.length = end - possible_start + 1
            return True
        return False



RequestPacket.register_element("IDENTIFIER", "initializing character identifying command", start_position=0, length=1)
RequestPacket.register_element("DATA", "data contained in packet", start_position=1, end_position=-2)
RequestPacket.register_element("TERMINATOR", "terminating characters", start_position=-2, length=2)

ResponsePacket = RequestPacket



from distutils.core import setup

setup(name = "pydcpf",
      version = "1.3",
      description = "Python device communications protocol framework",
      long_description = "A framework for rapid implementation of communications protocols in Python focusing on speed, modular flexibility and extensibility.",
      url="https://github.com/smartass101/pydcpf",
      packages = ["pydcpf", "pydcpf.interfaces", "pydcpf.appliances", "pydcpf.protocols"],
      author='Ondrej Grover',
      author_email='ondrej.grover@gmail.com',
      requires = ["pyserial"],
      classifiers = [
          "Development Status :: 5 - Production/Stable",
          "Intended Audience :: Science/Research",
          "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
          "Natural Language :: English",
          "Operating System :: OS Independent",
          "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
          ],
        
      )



#Python device communications protocol framework (pydcpf)
#Copyright (C) 2013  Ondřej Grover
#
#pydcpf is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#pydcpf is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with pydcpf.  If not, see <http://www.gnu.org/licenses/>.
"""This module provides the core class of the package: :class:`Device`"""

from types import ModuleType


class Device(object):
    """Core class used for communication with a device

    Uses a subclass of :class:`interfaces.base.Interface` class for data transmission and sublcasses of :class:`protocols.base.RequestPacket` and :class:`protocols.base.ResponsePacket` for data encoding and decoding respectively.
    """


    def __init__(self, address, protocol_module, interface_module=None, timeout=1.0, send_byte_count=0, receive_byte_count=8192, connect=True, serve=False, **interface_kwargs):
        """Initialize the device

        Parameters
        ----------
        address : interface dependant
            The address of the physical device used to create the underlying interface
            if the *interface_module* is not specified, pydcpf.interfaces.serial module is used if address is a str or int referencing a serial port (e.g. '/dev/ttyUSB0', '/dev/ttyS0' or a number of the serial/COMM port) or the pydcpf.interfaces.socket module is used if address is a tuple of (str, int) (e.g. ('192.168.1.254', 10001)) for Interface creation.
        protocol_module : str or module
            From where to load the packet classes, the module must provide a ResponsePacket and RequestPacket class (for many protocols they are the same)
        interface_module : str or module, optional
            From where to load the Interface class
            If not specified, the module is guessed from the *address* parameter
        timeout : float, optional
            timeout for connecting, reading and writing
            if 0, no timeout is set
            the error raised on timout depends on the implementation of the interface module
        send_byte_count : int or None, optional
            Size of byte chunks to send at once
            If 0, the bytes will not be split into chunks
        receive_byte_count : int or None, optional
            Size of byte chunks to receive at once
            If 0, the bytes will not be split into chunks
        connect : bool, optional
            If True, connect immediately after initialization
        serve : bool, optional
            If True, this device is expected to wait for a connection which may modify the way the device connects
        """
        self.data_buffer = bytearray()
        self.serve = serve
        self.address = address
        self.send_byte_count = send_byte_count
        self.receive_byte_count = receive_byte_count
        if not isinstance(protocol_module, ModuleType):
            protocol_module = __import__(protocol_module, fromlist=[''])
        self.protocol = protocol_module
        self._request_buffer_packet = protocol_module.RequestPacket()
        if not isinstance(interface_module, ModuleType):
            if interface_module is None:
                if isinstance(address, (int, str)): #seems to be a serial device
                    interface_module = 'pydcpf.interfaces.serial_interface'
                elif isinstance(address, tuple) and len(address) == 2 and isinstance(address[0], str) and isinstance(address[1], int): #appears to be a socket address
                    interface_module = 'pydcpf.interfaces.socket_interface'
            interface_module = __import__(interface_module, fromlist=[''])
        self.interface = interface_module.Interface(timeout, **interface_kwargs)
        if connect:
            self.connect()
            
        
    def connect(self, address=None, serve=None):
        """Connect to the device, optionally override and set the Device.address ad Device.serve attributes (same form and meaning as in :meth:`Device.__init__`)"""
        if address is None:
            address = self.address
        else:
            self.address = address
        if serve is None:
            serve = self.serve
        else:
            self.serve = serve
        self.interface.connect(address, serve)


    def disconnect(self):
        """Disconnect from the device in such a way that it can be connected to again"""
        self.interface.disconnect(self.address, self.serve)
        

    def send_request_packet(self, packet, send_byte_count=None):
        """Send a request packet, optionally send *send_byte_count* size byte chunks at once

        Parameters
        ----------
        packet : RequestPacket
            packet to send
        send_byte_count : int, optional
            if specified, this will override the Device.send_byte_count attribute specified during initialization
        """
        if send_byte_count is None:
            send_byte_count = self.send_byte_count
        raw_packet = packet.raw_packet
        if send_byte_count == 0: #do not split the packet TODO possibly to loose checking, what about None or negative values?
            self.interface.send_data(raw_packet)
        else: #may split
            for delimiter in xrange(0, len(raw_packet), send_byte_count):
                self.interface.send_data(raw_packet[delimiter:delimiter + send_byte_count])

                
    def send_request(self, send_byte_count=None, **packet_parameters):
        """Send a request, optionally send *send_byte_count* size byte chunks at once

        Parameters
        ----------
        send_byte_count : int, optional
            if specified, this will override the Device.send_byte_count attribute specified during initialization
        **packet_parameters
            keyword arguments containing parameters for packet creation
        """
        self._request_buffer_packet.__init__(**packet_parameters)
        self.send_request_packet(self._request_buffer_packet, send_byte_count)
            


    def receive_response(self, receive_byte_count=None, **check_parameters):
        """Receive a response and return only data, but check the packet first

        Parameters
        ----------
        receive_byte_count : int, optional
            if specified, this will override the Device.receive_byte_count attribute specified during initialization
        **check_parameters
            keyword arguments containing parameters for packet validation

        Returns
        -------
        data : str or buffer
            the data contained within the received packet DATA attribute
        """
        packet = self.receive_response_packet(receive_byte_count)
        packet.check(**check_parameters)
        return packet.DATA


    def receive_response_packet(self, receive_byte_count=None, packet=None):
        """Receive a response and return the received packet, optionally read *receive_byte_count* sized byte chunks at once into the specified *packet*

        Parameters
        ----------
        receive_byte_count : int, optional
            if specified, this will override the Device.receive_byte_count attribute specified during initialization
        packet : ResponsePacket, optional
            If specified, the data will be read into this packet and then it will be returned

        Returns
        -------
        packet : ResponsePacket
            Packet created from the protocol module (specified during initialization) or the one passed to this method
        """
        if packet is None:
            packet = self.protocol.ResponsePacket()
        if receive_byte_count is None:
            receive_byte_count = self.receive_byte_count
        packet.raw_packet = self.data_buffer
        while not packet.find():
            packet.raw_packet.extend(self.interface.receive_data(receive_byte_count))
        self.data_buffer = packet.raw_packet[packet.start + packet.length:]
        return packet

    
    def query(self, send_byte_count=None, receive_byte_count=None, check_parameters=dict(), **packet_parameters):
        """Query the device

        Essentially just a wraper around :method:`Device.send_request` and
        :method:`Device.receive_response`
        
        """
        self.send_request(send_byte_count, **packet_parameters)
        return self.receive_response(receive_byte_count, **check_parameters)
    


class Server(object):
    """Serves Devices

    Has the ability to find out protocol they want to use for communication
    """
    pass



#Python device communications protocol framework (pydcpf)
#Copyright (C) 2013  Ondřej Grover
#
#pydcpf is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#pydcpf is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with pydcpf.  If not, see <http://www.gnu.org/licenses/>.
from .spinelbase import SpinelBasePacket, ACKError

__all__ = ["RequestPacket", "ResponsePacket"]



class Spinel66BasePacket(SpinelBasePacket):

    
    def check(self):
        if self.ACK != '0':
            raise ACKError(self)



Spinel66BasePacket.register_element('ADR', "Module address character", start_position=2, length=1)
Spinel66BasePacket.register_element('DATA', "Data contained in packet", start_position=4, end_position=-1)



class RequestPacket(Spinel66BasePacket):
    """RequestPacket class for the Spinel 66 protocol format"""

    
    def __init__(self, INST=None, ADR='$', DATA='', raw_packet=None ):
        if INST is not None or raw_packet is not None:
            super(RequestPacket, self).__init__(raw_packet=raw_packet, PRE='*', FRM=66, ADR=ADR, INST=INST, DATA=DATA, CR='\r')



RequestPacket.register_element('INST', "Device instruction code character", start_position=3, length=1)



class ResponsePacket(Spinel66BasePacket):
    """ResponsePacket class for the Spinel 66 protocol format"""

    
    def __init__(self, ACK=None, ADR='$', DATA='', raw_packet=None ):
        if ACK is not None or raw_packet is not None:
            super(ResponsePacket, self).__init__(raw_packet=raw_packet, PRE='*', FRM=66, ADR=ADR, ACK=ACK, DATA=DATA, CR='\r')



ResponsePacket.register_element('ACK', 'Acknowledgment code character', start_position=3, end_position=4)
